/***********************************************************************\
* NLED Text Editor Copyright 1987-2003, Evan Alexander Weaver           *
*                                                                       *
* This source code, and the associated executable programs obtained by  *
* compiling it, are published under the GNU General Public License. See *
* the accompanying file "nled.txt" for the details of this license and  *
* some other comments about NLED and this source code.                  *
*                                                                       *
* File: NLEDIT.C                          Date Last Modified: 12-Feb-92 *
*                                                                       *
* Various functions controlling editing of file in NLED.                *
* Functions are:                                                        *
*   blockupd() - updates block_xxx vars after given shift at given pos. *
*   delete() - deletes part of a character string                       *
*   dispchgcase() - changes case of current char if alphabetic          *
*   dispcpyblk() - copies block_xxx-marked block into current position  *
*   dispdelblk() - deletes block_xxx-marked block                       *
*   dispdelc() - deletes a character from the display                   *
*   dispinsc() - inserts a given character n times into the display     *
*   dispmovblk() - moves block_xxx-marked block to current position     *
*   disprepc() - replaces current character with given character        *
*   filltoend() - pads out current line with spaces                     *
*   insert() - inserts one string into another                          *
*   insertc() - inserts a character n times into a string               *
*   nolines() - determines the number of lines between two points       *
*   offedge() - scrolls screen so cursor is not on top or bottom        *
*                                                                       *
* 05-Oct-91: ANSI function headers instead of "old-style"               *
* 12-Feb-92: UNIX standardization                                       *
\***********************************************************************/

#include <stdio.h>
#include "nled.h"
#include "nledkeys.h"
#include "nledvdef.h"
#include "nledfdef.h"

/***********************************************************************\
* Updates the block markers assuming that 'delta' characters have been  *
* added or removed from the data at position 'pos'. The global len_ds   *
* should already have been updated.                                     *
\***********************************************************************/
void blockupd(INDEX pos, INDEX delta)
{
    if (block_beg > pos)
        block_beg = max(pos,block_beg + delta);
    if (block_end > pos)
        block_end = max(pos,block_end + delta);
    if (block_end > len_ds) block_end = len_ds;
}

/***********************************************************************\
* General string deletion utility. Deletes 'number' characters from     *
* 'string[]' starting at offset 'start' where '*plen' is the length of  *
* the string (which is updated)                                         *
\***********************************************************************/
void delete(int number, int start, char *string, int *plen)
{
     int i, j;

     if ((number <= 0) || (start >= *plen)) return;

     if (start < *plen - number) {
        *plen = *plen - number;
        j = (i = start) + number;
        while (i < *plen)
            string[i++] = string[j++];
     }
     else
        *plen = start;
}

/***********************************************************************\
* Changes case of current character if alphabetic                       *
* 22-Apr-88: New routine                                                *
\***********************************************************************/
int dispchgcase(void)
{
    int c, offset;

    if (scr_left + col_no < scr_llen[row_no]) {
        c = datget(abs_pos);
        if (c >= 'a' && c <= 'z') 
            offset = ('A' - 'a');
        else if (c >= 'A' && c <= 'Z')
            offset = ('a' - 'A');
        else
            offset = 0;
        if (offset)
            disprepc((char)(c + offset));
    }
    return (OK);
}

/***********************************************************************\
* Copies marked block to current position, if possible.                 *
\***********************************************************************/
int dispcpyblk(void)
{
    INDEX i,n,srow;
    int c,retcode;

    if (block_beg >= block_end) {
        dispmess(" No block marked",-1,FALSE);
        retcode = ERR;
    }
    else if ((block_beg < abs_pos) && (abs_pos < block_end)) {
        dispmess(" Cannot copy block into itself",-1,FALSE);
        retcode = ERR;
    }
    else {
        srow = row_no;
        dispmess(" Copying block...",17,TRUE);
        filltoend();
        n = block_end - block_beg;
        n = datinsn(' ',n,abs_pos);
        len_ds += n;
        if (block_beg >= abs_pos) {
            block_beg += n;
            block_end += n;
        }
        i = abs_pos;
        datfindseq(block_beg);
        while (i < abs_pos + n) {
            if (LF == (c = datgetseq())) nol_ds++;
            datset(c,i++);
        }
        dispcalc(row_no);
        draw_erow = num_rows - 2;
        setdraw();
        edit_flag = TRUE;
        retcode = OK;
        if (srow != row_no) offedge();
    }
    return(retcode);
}

/***********************************************************************\
* Deletes marked block, if possible, and moves cursor there.            *
\***********************************************************************/
int dispdelblk(void)
{
    INDEX srow;
    int retcode;

    if (block_end > block_beg) {
        srow = row_no;
        dispmess(" Deleting block...",18,TRUE);
        dispupd('r', nolines(abs_pos,block_beg));
        dispupd('c', block_beg - scr_loff[row_no] - scr_left - col_no);
        nol_ds -= nolines(block_beg,block_end);
        len_ds -= datdeln(block_end - block_beg,block_beg);
        edit_flag = TRUE;
        block_end = block_beg = 0;
        dispcalc(row_no);
        draw_erow = num_rows-2;
        setdraw();
        if (srow != row_no) offedge();
        retcode = OK;
    }
    else {
        dispmess(" No block marked",-1,FALSE);
        retcode = ERR;
    }
    return(retcode);
}

/**********************************************************************\
* Deletes character from current display position, If we are past the  *
* end of the line, we delete the LF & bring the next line up to this   *
* position.                                                            *
\**********************************************************************/
void dispdelc(void)
{
    int i;

    if (scr_left + col_no >= scr_llen[row_no]) {
        /* we are past the end of the line */

        /* nothing to do if we're past end of the file */
        if (scr_top + row_no + 1 >= nol_ds) return;

        /* insert blanks */
        filltoend();

        /* we will be deleting a LF (which is at abs_pos) */
        nol_ds--;
        draw_erow = num_rows - 2; /* make sure we redraw to end of screen */

        datdeln(1,abs_pos);
        len_ds--;
        dispcalc(row_no);
    }
    else {
        for (i = row_no + 1; i < num_rows - 1; i++)
            scr_loff[i] = scr_loff[i] - 1;
        scr_llen[row_no] = scr_llen[row_no] - 1;
        datdeln(1,abs_pos);
        len_ds--;
    }
    blockupd(abs_pos,-1);
    setdraw();      /* redraw at least from this point on on this line */
    edit_flag = TRUE;
}

/**********************************************************************\
* Inserts 'character' into the current display position 'count' times. *
* If the current position is past the end of the line, blanks are used *
* to pad out the line to the current position, except when a LF is     *
* the character being inserted, in which case it is inserted right at  *
* the end of the line.                                                 *
\**********************************************************************/
void dispinsc(char character, INDEX count)
{
    int i;

    /* do nothing if there is nothing to insert */
    if (count <= 0) return;

    if (character == LF) {
        /* if LF is being inserted, don't pad out line with blanks */
        count = datinsn(character,count,abs_pos);

        nol_ds += count;
        len_ds += count;

        dispcalc(row_no);

        draw_erow = num_rows - 2;
    }
    else {
        /* if we're beyond end of line, insert blanks to fill it up */
        filltoend();

        count = datinsn(character,count,abs_pos);
        len_ds += count;

        for (i = row_no + 1; i < num_rows - 1; i++)
            scr_loff[i] = scr_loff[i] + count;
        scr_llen[row_no] = scr_llen[row_no] + count;
    }

    blockupd(abs_pos,count);
    setdraw();
    edit_flag = TRUE;
}

/***********************************************************************\
* Moves marked block to current position, if possible. Cursor is moved  *
* to start of the block in its new position.                            *
* 03-Oct-88: Minor adjustment for optimizer bug.                        *
\***********************************************************************/
int dispmovblk(void)
{
    INDEX nextloff, oldbb, srow;
    int i;

    if (block_beg >= block_end) {
        dispmess(" No block marked",-1,FALSE);
        return (ERR);
    }
    if ((block_beg <= abs_pos) && (abs_pos <= block_end)) {
        dispmess(" Cannot move block into itself",-1,FALSE);
        return(ERR);
    }

    srow = row_no;
    dispmess(" Moving block...",16,TRUE);
    filltoend();
    datmove(block_beg, block_end, abs_pos);

    if (abs_pos < block_beg) {
        /* all changes were after current position, updating screen is easy */
        block_end = abs_pos + block_end - block_beg;
        block_beg = abs_pos;
        dispcalc(row_no);
        setdraw();
        draw_erow = num_rows - 2;
    }
    else {
        /* changes were before current position */
        oldbb = block_beg;
        block_beg += abs_pos - block_end;
        block_end = abs_pos;

        if (scr_loff[0] <= oldbb) {
            /* top of screen is before changes */
            dispcalc(0);
            dispupd('r',-nolines(block_beg, block_end));
        }
        else {
            /* scr_xxx are wrong for this row & all before. However,  */
            /* nextloff (below) is correct, though its parts are not! */
            nextloff = scr_loff[row_no];
            nextloff += scr_llen[row_no] + 1;
            for (i = row_no; i >= 0; i--)
                nextloff = (scr_loff[i] = nextloff - 1 -
                 (scr_llen[i] = datpllen(nextloff)));
            i = row_no;
            dispupd('r',-nolines(block_beg, block_end));
            i = min(scr_top, i-row_no);
            dispupd('t',-i);
            dispupd('r',i);
        }
        dispupd('c',block_beg - scr_loff[row_no] - scr_left - col_no);
        setdrawall();
    }
    if (srow != row_no) offedge();
    edit_flag = TRUE;
    return(OK);
}

/**********************************************************************\
* Replaces current position display position with 'character'. This    *
* requires insert mode if we are past the end of the line.             *
\**********************************************************************/
void disprepc(char character)
{
    if (scr_left + col_no >= scr_llen[row_no]) 
        dispinsc(character,1);
    else
        datset(character,scr_loff[row_no] + scr_left + col_no);
    setdraw();
    edit_flag = TRUE;
}

/***********************************************************************\
* Pads from the end of the line to the current position with blanks, if *
* necessary.                                                            *
\***********************************************************************/
void filltoend(void)
{
    INDEX n;

    if (0 < (n = scr_left + col_no - scr_llen[row_no])) {
        dispupd('c',-n);
        dispinsc(' ',n);
        dispupd('c',n);
    }
}

/***********************************************************************\
* General string insertion utility. Inserts 'inlen' characters starting *
* with '*instr' into 'string[]' starting at position 'start' where      *
* '*plen' is the number of characters in 'string[]' and 'maxlen' is the *
* maximum number of characters allowed for 'string[]'. Updates '*plen'. *
\***********************************************************************/
void insert(char *instr, int inlen, int start, char *string, int *plen,
 int maxlen)
{
    int i, j, end;

    if (inlen <= 0) return;
    if (inlen > maxlen - start)
        end = maxlen;
    else
        end = start + inlen;
    if (inlen > maxlen - *plen) 
        *plen = maxlen;
    else
        *plen = *plen + inlen;
    j = inlen + (i = *plen - inlen - 1);
    while (i >= start)
        string[j--] = string[i--];
    j = 0;
    i = start;
    while (i < end)
        string[i++] = instr[j++];
}

/***********************************************************************\
* General string insertion utility. Inserts 'character' 'inlen' times   *
* into 'string[]' starting at position 'start' where '*plen' is the     *
* number of characters in 'string[]' and 'maxlen' is the maximum number *
* of characters allowed for 'string[]'. '*plen'is updated.              *
\***********************************************************************/
void insertc(char character, int repeat, int start, char *string, int *plen,
 int maxlen)
{
    int i, j, end;

    if (repeat <= 0) return;

    if (repeat > maxlen - start)
        end = maxlen;
    else
        end = start + repeat;
    if (repeat > maxlen - *plen) 
        *plen = maxlen;
    else
        *plen = *plen + repeat;
    j = repeat + (i = *plen - repeat - 1);
    while (i >= start)
        string[j--] = string[i--];
    for (i = start; i < end; i++)
        string[i] = character;
}

/***********************************************************************\
* Returns the number of lines (+ for up, - for down) from position      *
* 'pos1' to position 'pos2'.                                            *
\***********************************************************************/
INDEX nolines(INDEX pos1, INDEX pos2)
{
    INDEX count = 0;

    if (pos1 < pos2) {
        datfindseq(pos1);
        do {
            if (LF == datgetseq()) count++;
        } while (++pos1 < pos2);
    }
    else if (pos1 > pos2) {
        datfindseq(pos2);
        do {
            if (LF == datgetseq()) count--;
        } while (pos1 > ++pos2);
    }
    return(count);
}

/***********************************************************************\
* Scrolls the screen and moves the cursor so that the cursor is not on  *
* either the top or the bottom line of the text area, if possible.      *
\***********************************************************************/
void offedge(void)
{
    if ((row_no == 0) && (scr_top > 0)) {
        dispupd('t',-1);
        dispupd('r',1);
    }
    else if ((row_no == num_rows - 2) && (1 + scr_top + row_no < nol_ds)) {
        dispupd('t',1);
        dispupd('r',-1);
    }
}
/*--------------------------END-OF-NLEDIT.C---------------------------*/
